#include "cthugha.h"
#include "interface.h"
#include "display.h"
#include "keys.h"
#include "imath.h"

InterfaceList::InterfaceList(InterfaceElement ** fixed, int nEl, OptionList * ol) : 
    Interface(NULL, 0), pos(nEl), optList(ol) {

    InterfaceElement ** el = new InterfaceElement*[nEl + ol->nEntries];

    // copy the fixed entries
    for(int i=0; i < nEl; i++)
	el[i] = fixed[i];

    // make the new entries
    for(int i=0; i < ol->nEntries; i++) {
	el[i+nEl] = new InterfaceElementOption("", ol->use[i]);
    }

    elements = el;
    nElements = nEl + ol->nEntries;
    nFixed = nEl;

    sel = 0;
    elements[sel]->selected = 1;
}

int InterfaceList::do_key(int key) {
    switch(key) {
    case CK_ENTER:
	if(sel >= nFixed) {
	    optList->use[sel-nFixed]->changeTo(1); // make it usable
	    optList->changeTo(sel - nFixed);	// make it active
	    setInterface(interfaceMain);	// and close this display
	    return 0;
	} else
	    return key;
    case ' ':
	if(sel >= nFixed) {
	    optList->use[sel-nFixed]->changeTo(1);	// make it usable
	    optList->changeTo(sel - nFixed);	// make it active
	    return 0;
	} else
	    return key;
    }
    return Interface::do_key(key);
}

void InterfaceList::display() {
    int l;
    // fixed elements (title, ...)
    for(l=0; l < nFixed; l++)		
	elements[l]->display(l);

    if( nElements == nFixed)
	return;
    
    // scrolling elements
    if( sel < pos) {
	pos = max(sel, nFixed);
    }
    if( (sel-size) >= pos) {
	pos = sel;
    }
    for(int i = 0; i < size; i++) {
	int s = i + pos;
	if ( (s >= nElements) || (s < nFixed) )
	    continue;

	char str1[128];
	sprintf(str1, "%c%s",
		(s == sel) ? '>' : ' ',
		optList->entries[s-nFixed]->name);

	char str2[128];
	sprintf(str2, "%s %3s%c",
		optList->entries[s-nFixed]->desc,
		optList->use[s-nFixed]->text(),
		(s == sel) ? '<' : ' ');

	char str3[128];
	sprintf(str3, "%40s", str2);

	display_print(str3, (i+l), 'l', (s==sel) ? TEXT_COLOR_HIGHLIGHT : TEXT_COLOR_NORMAL);
	display_print(str1, (i+l), 'l', (s==sel) ? TEXT_COLOR_HIGHLIGHT : TEXT_COLOR_NORMAL);
    }
}

int InterfaceList::size = 6;




