#include "cthugha.h"
#include "options.h"
#include "display.h"

OptionOnOff options_save("save", 0);	// save options (and buffer) when leaving cthugha


Option * Option::feature = NULL;
int Option::nFeatures = 0;

#define MAX_HISTORY 128


//
// Option
//
Option::Option(const char * n, int hist) :
    value(0), 
    history(hist),
    name(n) {

    if(history) {
	// keep in linked list
	next = feature;
	feature = this;

	history = 1;
	oldValues = new int[MAX_HISTORY];

	hot = new int[10];
	for(int i=0; i < 10; i++)
	    hot[i] = 0;

	nFeatures ++;
    } else {
	hot = NULL;
	oldValues = NULL;
	next = NULL;
    }
}

//
// save the current state
//
void Option::doSave() {
    if(oldValues == NULL)
	return;

    if(history >= MAX_HISTORY) {
	memcpy(oldValues, oldValues+1, sizeof(int) * (MAX_HISTORY-1) );
	history --;
    }
    oldValues[history] = value;
    history ++;
}

//
// get back the last state
//
void Option::doRestore() {
    if(oldValues == NULL)
	return;

    if(history > 0) {
	history --;
	changeTo( int(oldValues[history]) );
    }
}

//
// save the state of all options
//
void Option::save() {
    for(Option * o = feature; o != NULL; o = o->next) 
	o->doSave();
}

//
// get back the state of all options
//
void Option::restore() {
    for(Option * o = feature; o != NULL; o = o->next) 
	o->doRestore();
}

//
// save to hotkey position
//
void Option::save(int to) {
    if( (to < 0) || (to >= 10))
	return;

    for(Option * o = feature; o != NULL; o = o->next) 
	if(o->hot)
	    o->hot[to] = o->value;
}

//
// get back from hotkey
//
void Option::restore(int from) {
    if( (from < 0) || (from >= 10))
	return;

    for(Option * o = feature; o != NULL; o = o->next)
	if(o->hot)
	    o->changeTo( o->hot[from]);
}

//
// default handler for changeTo with names
//
void Option::changeTo(const char * to) {
    char * pos;

    /* try to read as number */
    int tvalue = strtol(to, &pos, 0);
    if (to == pos ) {		// not a number
	value = 0;
	return;
    }

    value = tvalue;

    changeTo(atoi(to));
}

//
// change randomly one of the features
//
void Option::changeOne() {

    if(feature == NULL)
	return;

    int n = rand() % nFeatures;
    Option * o = feature;
    while( (n > 0) && (o != NULL) ) {
	o = o->next;
	n--;
    }

    if(o == NULL)	// just to be sure
	return;

    save();

    o->changeTo(CHANGE_RANDOM);
}

//
// change all the features
//
void Option::changeAll() {

    save();

    for(Option * o = feature; o != NULL; o = o->next) {
	o->changeTo(CHANGE_RANDOM); 
    }
}

//
// read and write hotkeys settings to ini files
//
void Option::getHotIni() {
    for(int i=0; i < 10; i++) {
	for(Option * o = feature; o != NULL; o = o->next) {
	    char str[512], val[512];
	    sprintf(str, "hot.%d.", i);
	    strncat(str, o->name, 512);
	    if(! getini(str, val)) {
		o->changeTo(val);
		o->hot[i] = o->value;
	    }
	}
    }
}
void Option::putHotIni() {
    for(int i=0; i < 10; i++) {
	for(Option * o = feature; o != NULL; o = o->next) {
	    char str[512];
	    sprintf(str, "hot.%d.", i);
	    strncat(str, o->name, 512);
	    putini(str, o->text(o->hot[i]) );
	}
    }
}



void OptionOnOff::changeTo(const char * to) {

    if ( ! strncasecmp("yes", to, 3) )
	value = 1;
    else if ( ! strncasecmp("on", to, 2) )
	value = 1;
    else if ( ! strncasecmp("1", to, 1) )
	value = 1;
    else if ( ! strncasecmp("no", to, 2) )
	value = 0;
    else if ( ! strncasecmp("off", to, 3) )
	value = 0;
    else if ( ! strncasecmp("0", to, 1) )
	value = 0;
    else {
	printfe("Illegal yes/no-value `%s'.\n", to);
    }
}

    






/*
 * Set first options in use
 */
#include "waves.h"
#include "translate.h"
#include "action.h"
int init_startup() {
    /* set screen-function */
    screen.changeTo(screen_first);

    /* make sure we have at least some a palette */
    palette.changeTo(CHANGE_RANDOM);

    /* use first pcx if specified */
    if( pcx_first[0] != '\0' ) { 
        /* select first pcx-file */
	pcx.changeTo(pcx_first);
        if ( pcx.current() != NULL) {
	    show_pcx(1,0);		// initialize the buffers
//	    active_palette =		/* and set corresponding palette */
	    // 	        feature_change(pcx_palettes[pcxs.current],
	    //		       &palettes)->data;
	}
    } else {
	pcx.changeTo(CHANGE_RANDOM);
    }

    /* set palette */
    if( palette_first[0] != '\0') {
	palette.changeTo(palette_first);
    } 

    /* set the features to start with */
    wave.changeTo(wave_first);
    object.changeTo(object_first);
    translate.changeTo(translate_first);
    flame.changeTo(flame_first);

    return 0;
}

/*
 * save startup-section if --save was given
 */
int exit_startup() {
    if ( int(options_save) ) {
	write_ini();
	save_buffer();
    }
    return 0;
}
