#include "cthugha.h"
#include "sound.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <math.h>
#include <sys/ioctl.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <signal.h>
#include <memory.h>

#include <sys/types.h>
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif


#if TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif



class sound_Communicator {
public:
    int sample_rate;
    int channels;
    int sample_size;
    int method;
    int fragments;
    int fragment_size;

    int update;
    int done_update;
    int stop;
};
sound_Communicator * sound_communicator = NULL;


void sig_tty_parent(int ) {
    printfv(0, "Stopping in child...\n");

    kill(getppid(), SIGTSTP);

    sound_communicator->stop = 1;
}



int sound_comm_read();
int sound_communicate(int to_child) {
    if(sound_communicator == NULL)
	return 0;

    sound_communicator->method        = int(sound_method);
    sound_communicator->sample_rate   = int(sound_sample_rate);
    sound_communicator->channels      = int(sound_channels);
    sound_communicator->sample_size   = int(sound_sample_size);
    sound_communicator->fragments     = int(sound_fragments);
    sound_communicator->fragment_size = int(sound_fragment_size);
    if(to_child) {
	sound_communicator->update = 1;
	sound_communicator->done_update = 0;
    } else {
	sound_communicator->update = 0;
	sound_communicator->done_update = 1;
    }
    return 0;
}
int sound_comm_read() {
    if(sound_communicator == NULL)
	return 0;

    sound_method.changeTo( sound_communicator->method);
    sound_sample_rate.changeTo( sound_communicator->sample_rate);
    sound_channels.changeTo( sound_communicator->channels);
    sound_sample_size.changeTo( sound_communicator->sample_size);
    sound_fragments.changeTo( sound_communicator->fragments);
    sound_fragment_size.changeTo( sound_communicator->fragment_size);
    return 0;
}

/*
 * fork the sound reading process
 */
SoundDeviceFork::SoundDeviceFork() {
    printfv(3,"    starting sound reading process\n");
    if( (sound_shm_key = shmget(IPC_PRIVATE, sound_bsize*2 + 
				sizeof(sound_Communicator), 
				IPC_CREAT | 0777) ) == -1) {
	printfee("Can not create shared memory segment.");
	sound_fork.changeTo(0);
	error = 1;
	return;
    }
    if( (sound_shared=shmat(sound_shm_key, 0, 0)) == (void*)-1) {
	printfee("Can not attach shared memory segment.");
	sound_fork.changeTo(0);
	error = 1;
	return;
    }

    sound_communicator = (sound_Communicator*)sound_shared;
    sound_communicate(1);
    sound_communicator->stop = 0;

    sound_shared += sizeof(sound_Communicator);

    switch( sound_child = fork() ) {
    case -1:		/* error */
	printfee("Can not fork sound child.");
	shmdt((char*)sound_data);

	error = 1;
	return;

    case 0:		/* child */
	sound_fork.changeTo(0);		/* don't fork again */
	if(init_sound()) {		/* open sound device */
	    abort();
	}
	sound_communicate(0);		// and send the new values back

	/* set signal handler */
	signal(SIGTSTP, sig_tty_parent);	/* react to ^Z */

	sound_data = (char2*)sound_shared;

	do {
	    soundDevice->read();
	    if(sound_communicator->update) {
		sound_comm_read();		// this only sets the values
		soundDevice->update();		// really update device
		sound_communicate(0);		// and send the new value back
	    }
	} while( !sound_communicator->stop);
	printfv(3, "    closing sound reading child.\n");
	abort();

    default: {
	    sleep(1);	// to make the text output nicer
	    nice(10);	// be nice, so that the sound reading proc gets more priority
	}
    }
}

void SoundDeviceFork::read() {
    memcpy(sound_data, sound_shared, sound_bsize*2 );
    if(sound_communicator->done_update) {
	sound_comm_read();
	sound_communicator->done_update = 0;
    }
}

void SoundDeviceFork::update() {
    sound_communicate(1);
}



/*
 * kill the sound reading process
 */
SoundDeviceFork::~SoundDeviceFork() {
    printfv(3,"    stopping sound reading process.\n");
    sound_communicator->stop = 1;

    if(sound_child >= 0) 
	if( waitpid(sound_child, NULL, WNOHANG) == -1) {
	    // first wait failed, wait a short time
	    sleep(2);
	    if( waitpid(sound_child, NULL, WNOHANG) == -1) {
		// second wait failed, now be brutal
		kill(SIGKILL, sound_child);
		waitpid(sound_child, NULL, 0);
	    }
	}

    shmdt((char*)sound_shared);
    shmctl(sound_shm_key, IPC_RMID, 0);
}

