#include "cthugha.h"
#include "display.h"
#include "options.h"
#include "keys.h"
#include "action.h"
#include "information.h"
#include "sound.h"
#include "translate.h"
#include "interface.h"
#include "disp-sys.h"
#include "imath.h"
#include "cth_buffer.h"

#include <math.h>

/* possible display-function */
int screen_up();	int screen_down();	int screen_2hor();
int screen_r2hor();	int screen_4hor();	int screen_2verd();
int screen_r2verd();	int screen_4kal();	int screen_hfield();
int screen_roll();	int screen_zick();	int screen_bent();
int screen_plate();

class ScreenEntry : public  OptionListEntry {
public:
    int (*screen)();

    ScreenEntry(int (*f)(), const char * name, const char * desc) :
	OptionListEntry(name, desc), screen(f) {
    }

    int operator()() {
	return (*screen)();
    }
};

static OptionListEntry * _screens[] = {
    new ScreenEntry(screen_up,		"Up",		"Up Display" ),		// 0
    new ScreenEntry(screen_down,	"Down",		"Upside Down"),		// 1
    new ScreenEntry(screen_2hor,	"2hor",		"Hor. Split out"),	// 2
    new ScreenEntry(screen_r2hor,	"r2hor",	"Hor. Split in"),	// 3
    new ScreenEntry(screen_4hor,	"4hor",		"Kaleidoscope"),	// 4
    new ScreenEntry(screen_2verd,	"2verd",	"90deg rot. mirror"),	// 5
    new ScreenEntry(screen_r2verd,	"r2verd",	"90deg rot. mirror II"),// 6
    new ScreenEntry(screen_4kal,	"4kal",		"90deg Kaleidoscope"),	// 7
    new ScreenEntry(screen_hfield,	"hfield",	"Heightfield"),		// 8
    new ScreenEntry(screen_roll,	"roll",		"Roll around x-axis"),	// 9
    new ScreenEntry(screen_zick,	"zick",		"Zick Zack"),		// 10
    new ScreenEntry(screen_bent,	"bent",		"A bending plane"),	// 11
    new ScreenEntry(screen_plate,	"plate",	"A rotating plate")	// 12
};
static OptionOnOff * _screensUse[] = {
    new OptionOnOff("",1), // 0
    new OptionOnOff("",1), // 1
    new OptionOnOff("",1), // 2
    new OptionOnOff("",1), // 3
    new OptionOnOff("",1), // 4
    new OptionOnOff("",1), // 5
    new OptionOnOff("",1), // 6
    new OptionOnOff("",1), // 7
    new OptionOnOff("",1), // 8
    new OptionOnOff("",1), // 9
    new OptionOnOff("",0), // 10  don't use Zick Zack.
    new OptionOnOff("",1), // 11
    new OptionOnOff("",1)  // 12
};
OptionList screen("display", _screens, _screensUse,  sizeof(_screens) / sizeof(OptionList*));

xy screen_sizes[] = {
    {1,1},	/* up */	
    {1,1},	/* down */
    {1,1},	/* 2hor */	
    {1,1},	/* r2hor */
    {1,1},	/* 4hor */	
    {1,1},	/* 2verd */
    {1,1},	/* r2verd */	
    {1,1},	/* 4 kal */
    {2,2},      /* hfield */     
    {1,1},	/* rooll */
    {1,1},      /* zick */
    {2,2},	/* bent */
    {2,2},	/* plate */
};


OptionAllow rgb("rgb", NULL, OptionAllow::disabled);

xy disp_size = {0,0};				/* size of drawing area */
xy disp_one_size = {0,0};
xy draw_size = {0,0};				/* size of the drawn image */

OptionOnOff tileHoriz("tile-horizontal", 1);
OptionOnOff tileVert("tile-vertical", 1);

OptionAllow mirrorHoriz("mirror-horizontal", NULL, 0);
OptionAllow mirrorVert("mirror-vertical", NULL, 0);

int bytes_per_line = 0;

unsigned char * active_buffer;			/* Next on screen */
unsigned char * passive_buffer;			/* right now on screen */

char screen_first[256] = "";			/* Start with this scrn-fkt */

int display_frames = 0;				/* counter for speed-tests */
time_t display_start;

int display_direct = 0;				/* draw directly to screen */

int display_mode = 0;				/* graphic mode */

int bpp = 8;					/* bits per pixel */
int bypp = 1;					/* bytes per pixel */

unsigned char * display_mem = NULL;		/* memory for screen-fkt to draw to */

enum draw_mode_t draw_mode = DM_direct;		/* how drawing is done */

int red_mask, green_mask, blue_mask;
int red_shift, green_shift, blue_shift;

Interface * interfaceScreen;
static InterfaceElement * interfaceScreenElements[] = {
    new InterfaceElementTitle("Display"),
    new InterfaceElementText("---------------------------------------"),
};
static int nInterfaceScreenElements = sizeof(interfaceScreenElements)/sizeof(InterfaceElement*);

Interface * interfacePalette;
static InterfaceElement * interfacePaletteElements[] = {
    new InterfaceElementTitle("Palette"),
    new InterfaceElementText("---------------------------------------"),
};
static int nInterfacePaletteElements = sizeof(interfacePaletteElements)/sizeof(InterfaceElement*);


static xy disp_nr_tiles = {0,0};			/* nr of tiles to do */

static unsigned char * display_buffer = NULL;		// draw into this when not DM_direct
static int buffer_width = 0;

/*
 * Initialization of display 
 *  - load palettes and PCX 
 *  - allocate buffer
 *  - does not switch to graph mode
 */
int init_display() {
    if( init_display_sys() )
	return 1;

    /* The buffer has a border of 3 lines on top on bottom. These
       lines are set black before the 'flame' function is called.
       The border is not displayed to the screen
       */
    active_buffer  = new unsigned char[BUFF_SIZE+6*BUFF_WIDTH] + 3*BUFF_WIDTH;
    passive_buffer = new unsigned char[BUFF_SIZE+6*BUFF_WIDTH] + 3*BUFF_WIDTH;

    /* clear buffers */
    memset(active_buffer, 0, BUFF_SIZE);
    memset(passive_buffer, 0, BUFF_SIZE);

    display_buffer = new unsigned char[BUFF_SIZE * 4 * 2];


    /* load the palettes (internal and external) */
    if( load_palettes() )
	return 1;
		
    /* load the pcx-files */
    if ( init_pcx() )
	return 1;

    /* check if any palettes in use */
    if ( palette.getNEntries() == 0) {
	printfe("No palettes specified.\n");
	return 1;
    }

    interfaceScreen = new InterfaceList(interfaceScreenElements, 
					nInterfaceScreenElements, &screen);
    interfacePalette = new InterfaceList(interfacePaletteElements, 
					 nInterfacePaletteElements, &palette);

    return 0;
}

/*
 * Clean upt display
 */
int exit_display() {
    exit_graph_mode();
    return 0;
}


/*
 * Turn on graph-mode
 *   - turnon graph mode
 *   - allocate XImage
 */
int init_graph_mode() {

    printfv(3, "Initializing graphics mode...\n");

    if( init_graph_mode_sys() )
	return 1;

    printfv(3, "  display size: %dx%d\n", disp_size.x, disp_size.y);
    printfv(3, "  bytes/line  : %d\n", bytes_per_line);
    printfv(3, "  bytes/pixel : %d\n", bypp);

    display_start = gettime();
	
    return 0;
}

/*
 * Switch back to text mode
 */
int exit_graph_mode() {

    printfv(3, "Closing graphics mode...\n");

    exit_graph_mode_sys();
    exit_palettes();

    return 0;
}



/*****************************************************************************/

/*
 * exchange active and passive buffer
 */
void swap_buffers() {
    unsigned char * tmp = active_buffer; active_buffer = passive_buffer; passive_buffer = tmp;
}



/*
 * calculate information for tile and offset 
 */
int update_draw_size() {
    xy s = screen_sizes[screen.currentN()];

    disp_nr_tiles.x = int(tileHoriz) ? (disp_size.x/BUFF_WIDTH/s.x) : 1;
    disp_nr_tiles.y = int(tileVert) ? (disp_size.y/BUFF_HEIGHT/s.y) : 1;

    disp_one_size.x = s.x * BUFF_WIDTH;
    disp_one_size.y = s.y * BUFF_HEIGHT;

    draw_size.x = disp_nr_tiles.x * disp_one_size.x;
    draw_size.y = disp_nr_tiles.y * disp_one_size.y;

    if(draw_mode == DM_direct)
	buffer_width = bytes_per_line;
    else
	buffer_width = int(mirrorHoriz) ? disp_one_size.x * 2 : disp_one_size.x;

    return 0;
}

/*
 * clear border around display if necessary 
 */
static int clear_border() {
    static int display_clear = 1;				/* clear screen */

    if ( display_text_on_screen || display_clear ) {
	int bwidth = SCREEN_OFFSET_X;		/* width of left/right border */
	int bheight = SCREEN_OFFSET_Y;		/* height of upper/lower border */

	/* upper border */
	disp_clear_box_sys(0,0, disp_size.x, bheight);
	/* left border */
	disp_clear_box_sys(0, bheight, bwidth, draw_size.y);
	/* right border */
	disp_clear_box_sys(disp_size.x - bwidth, bheight, bwidth, draw_size.y);
	/* lower border */
	disp_clear_box_sys(0,disp_size.y-bheight, disp_size.x, bheight);

	/* if now text is on screen, we have to clean in the next iteration,
	   (maybe the text gets removed */
	display_clear = (display_text_on_screen) ? 1 : 0;
	return 1;
    }
    return 0;
}


//
// do the horizontal mirroring
// this is done in the temporary buffer
//
static void mirror_horiz() {
    if( (disp_nr_tiles.x > 1) && int(mirrorHoriz) ) {
	unsigned char * src = display_mem;
	unsigned char * dst = display_mem + (2*disp_one_size.x-1);
    
	for(int i=disp_one_size.y; i != 0; i--) {
	    for(int j=disp_one_size.x; j != 0; j--) {
		* dst = * src;
		src ++; dst --;
	    }
	    src += buffer_width - disp_one_size.x;
	    dst += buffer_width + disp_one_size.x;
	}
    }
}

//
// do the vertical mirroring
// this is done in the real display screen
//
static void mirror_vert() {
    if( (disp_nr_tiles.y > 1) && int(mirrorVert) ) {
	unsigned char * src = display_mem;
	unsigned char * dst = display_mem + (2*disp_one_size.y-1) * bytes_per_line;
    
	for(int i = disp_one_size.y; i != 0; i--) {
	    memcpy(dst, src, draw_size.x * bypp);
	    src += bytes_per_line;
	    dst -= bytes_per_line;
	}
    }
}

int tile_buffer() {
    int i;

    if( (disp_nr_tiles.x > 1) && int(mirrorHoriz) ) {		/* mirror */
	
	for(i=1; i < disp_nr_tiles.x/2; i++)
	    disp_copy_box_sys(SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			      disp_one_size.x*2, disp_one_size.y,
			      SCREEN_OFFSET_X + i*2*disp_one_size.x,
			      SCREEN_OFFSET_Y);
	/* special case for odd number of tiles */
	if( disp_nr_tiles.x & 1) 
	    disp_copy_box_sys(SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			      disp_one_size.x, disp_one_size.y,
			      SCREEN_OFFSET_X + i*2*disp_one_size.x,
			      SCREEN_OFFSET_Y);

    } else {							/* tile only */
	for(i=1; i < disp_nr_tiles.x; i++)
	    disp_copy_box_sys(SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			      disp_one_size.x, disp_one_size.y,
			      SCREEN_OFFSET_X + i*disp_one_size.x,
			      SCREEN_OFFSET_Y);
    }

    if( (disp_nr_tiles.y > 1) && int(mirrorVert) ) {		/* mirror */

	for(i=1; i < disp_nr_tiles.y/2; i++)
	    disp_copy_box_sys(SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			      draw_size.x, 2*disp_one_size.y,
			      SCREEN_OFFSET_X, 
			      SCREEN_OFFSET_Y + i*2*disp_one_size.y);
			      
	/* special case for odd number of tiles */
	if(disp_nr_tiles.y & 1) 
	    disp_copy_box_sys(SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			      draw_size.x, disp_one_size.y,
			      SCREEN_OFFSET_X, 
			      SCREEN_OFFSET_Y + i*2*disp_one_size.y);
			      
    } else {							/* tile only */
	for(i=1; i < disp_nr_tiles.y; i++)
	    disp_copy_box_sys(SCREEN_OFFSET_X, SCREEN_OFFSET_Y,
			      draw_size.x, disp_one_size.y,
			      SCREEN_OFFSET_X, 
			      SCREEN_OFFSET_Y + i*disp_one_size.y);
    }
    return 0;
}


//
// copy from the temporary buffer to the screen 
// this converts to the correct number of bytes
//
void buffer2Screen(unsigned char * buffer, unsigned char * screen) {

    int width;
    if( (disp_nr_tiles.x > 1) && int(mirrorHoriz) )
	width = disp_one_size.x * 2;
    else
	width = disp_one_size.x;

    switch(draw_mode) {								
    case DM_mapped1: case DM_tmp_mapped:
	for(int i=disp_one_size.y; i != 0; i--) {
	    unsigned long * scrn = (unsigned long*)screen;
	    unsigned char * buff = buffer;
	    for(int j=width/4; j != 0; j--) {
		unsigned long a;
		a  = bitmap_colors0[*buff]; buff++;
		a |= bitmap_colors1[*buff]; buff++;
		a |= bitmap_colors2[*buff]; buff++;
		a |= bitmap_colors3[*buff]; buff++;
		*scrn = a;
		scrn ++;
	    }
	    screen += bytes_per_line;
	    buffer += buffer_width;
	}
	break;
    case DM_mapped2:
	for(int i=disp_one_size.y; i != 0; i--) {
	    unsigned long * scrn = (unsigned long*)screen;
	    unsigned char * buff = buffer;
	    for(int j=width/2; j != 0; j--) {
		unsigned long a;
		a  = bitmap_colors0[*buff]; buff++;
		a |= bitmap_colors1[*buff]; buff++;
		*scrn = a;
		scrn ++;
	    }
	    screen += bytes_per_line;
	    buffer += buffer_width;
	}
	break;
    case DM_mapped3:
	for(int i=disp_one_size.y; i != 0; i--) {
	    unsigned long * scrn = (unsigned long*)screen;
	    unsigned char * buff = buffer;
	    for(int j=width/3; j != 0; j--) {
		unsigned long a, b;
		a = bitmap_colors0[*buff]; buff++;
		b = bitmap_colors0[*buff]; buff++;
		*scrn = a | ((b & 0xff) << 24); scrn ++;
		a = b >> 8;
		
		b = bitmap_colors0[*buff]; buff++;
		*scrn = a | ((b & 0xffff) << 16); scrn ++;
		a = b >> 16;

		b = bitmap_colors0[*buff]; buff++;
		*scrn = a | (b << 8); scrn ++;
	    }
	    screen += bytes_per_line;
	    buffer += buffer_width;
	}
	break;
    case DM_mapped4:
	for(int i=disp_one_size.y; i != 0; i--) {
	    unsigned long * scrn = (unsigned long*)screen;
	    unsigned char * buff = buffer;
	    for(int j=width; j != 0; j--) {
		*scrn = bitmap_colors0[*buff];
		buff ++;
		scrn ++;
	    }
	    screen += bytes_per_line;
	    buffer += buffer_width;
	}
	break;
    case DM_direct:
	;
    }
}

/*
 * Bring the active Buffer to screen.
 */
int display() {
    int clear;

    unsigned char * display_base = pre_draw_sys();

    do {
	update_draw_size();
	display_mem = (draw_mode == DM_direct) ? display_base + SCREEN_OFFSET : display_buffer;
    } while( screen() );			/* draw the buffer */

    mirror_horiz();

    if(draw_mode != DM_direct) {		// go back to real screen
	buffer2Screen(display_buffer, display_base + SCREEN_OFFSET);
	display_mem = display_base + SCREEN_OFFSET;
    }

    mirror_vert();

    tile_buffer();
    clear = clear_border();
    
    display_text_erase();
    currentInterface->display();		// print the text of the current interface
    errors.display();				// and the error messages
    display_text_refresh();

    disp_copy_sys(clear);

    swap_buffers();
    display_frames ++;
    
    return 0;
}




/*****************************************************************************
 * Screen-functions
 ****************************************************************************/

/*
 * Draw a permutation of the lines.
 * 
 * A lot of the display functions are now based on this.
 */
static unsigned char * perm_lines[MAX_BUFF_HEIGHT];

void screen_perm(void) {
    unsigned char * scrn = display_mem;
    unsigned char ** perm = perm_lines;
    int i;

    for(i=BUFF_HEIGHT; i != 0; i--) {
        memcpy(scrn, *perm, BUFF_WIDTH);
	scrn += buffer_width;
        perm ++;
    }
}

#define PERM_ADDR(i)	( active_buffer + (i) * BUFF_WIDTH )


int screen_up() {
    int i;
    for(i=0; i < BUFF_HEIGHT; i++)
	perm_lines[i] = PERM_ADDR(i);
		
    screen_perm();
    return 0;
}


int screen_down() {
    int i;
    for(i=0; i < BUFF_HEIGHT; i++)
	perm_lines[i] = PERM_ADDR(BUFF_HEIGHT - i - 1);
    
    screen_perm();
    return 0;
}

int screen_2hor() {
    int i;
    for(i=0; i < BUFF_HEIGHT/2; i++) {
	/* lower half of buffer maps to upper half of screen */
	perm_lines[i] = PERM_ADDR(BUFF_HEIGHT/2 + i);
	/* lower half of buffer get turned around to lower half of screen*/
	perm_lines[i+BUFF_HEIGHT/2] = PERM_ADDR(BUFF_HEIGHT - i - 1);
    }
	
    screen_perm();
    return 0;
}

		
int screen_r2hor() {
    int i;
    for(i=0; i < BUFF_HEIGHT/2; i++) {
	/* lower half of buffer get turned around to upper half of screen*/
	perm_lines[i] = PERM_ADDR(BUFF_HEIGHT - i - 1);
	/* lower half of buffer maps to lower half of screen */
	perm_lines[i+BUFF_HEIGHT/2] = PERM_ADDR(BUFF_HEIGHT/2 + i);
    }

    screen_perm();
    return 0;
}


int screen_4hor() {
    unsigned char * tmp, * scrn;
    int x,y;
	
    /* upper half of screen */	
    tmp = active_buffer + BUFF_SIZE/2;
    scrn = display_mem;
    for(y=BUFF_HEIGHT/2; y != 0; y--) {

	/* left half */
	memcpy( scrn, tmp, BUFF_WIDTH/2);
	scrn += BUFF_WIDTH/2;
	tmp += BUFF_WIDTH/2;
				
	/* right half */	
	for(x=BUFF_WIDTH/2; x != 0; x--) {
	    *scrn = *tmp;
	    scrn ++;
	    tmp --;		
	}			
	tmp += BUFF_WIDTH;	
	scrn += buffer_width - BUFF_WIDTH;
    }

    /* lower half of screen */
    tmp = active_buffer + BUFF_WIDTH*BUFF_HEIGHT - BUFF_WIDTH;
    scrn = display_mem + buffer_width * BUFF_HEIGHT/2;
    for(y= BUFF_HEIGHT/2; y != 0; y--) {
					
	/* left half */			
	memcpy(scrn, tmp, BUFF_WIDTH/2);
	scrn += BUFF_WIDTH/2;
	tmp += BUFF_WIDTH/2;	
				
	/* right half */	
	for(x=BUFF_WIDTH/2; x != 0; x--) {
	    *scrn = *tmp;		
	    scrn ++;
	    tmp --;			
	}				
	tmp -= BUFF_WIDTH;		
	scrn += buffer_width - BUFF_WIDTH;
    }

    return 0;
}


int screen_2verd() {
    if( BUFF_WIDTH/2 <= BUFF_HEIGHT) {
	int x,y;
	unsigned char * tmp = active_buffer;
	unsigned char * scrn = display_mem;
	for(y=BUFF_HEIGHT; y != 0; y--) {
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;			
		scrn ++;
		tmp += BUFF_WIDTH;		
	    }				
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;		
		scrn ++;
		tmp -= BUFF_WIDTH;		
	    }				
	    tmp ++;				
	    scrn += buffer_width - BUFF_WIDTH;
	}
	
    } else {
	screen.changeTo(CHANGE_NEXT);
	return 1;
    }
    return 0;
}


int screen_r2verd() {
    if( BUFF_WIDTH/2 <= BUFF_HEIGHT) {
	int x,y;
	unsigned char * tmp = active_buffer;
	unsigned char * scrn = display_mem + buffer_width*(BUFF_HEIGHT-1) + (BUFF_WIDTH-1);
	for(y=BUFF_HEIGHT; y != 0; y--) {		
	    for(x=BUFF_WIDTH/2; x != 0; x--) {	
		*scrn = *tmp;			
		scrn --;
		tmp += BUFF_WIDTH;			
	    }					
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;
		scrn --;
		tmp -= BUFF_WIDTH;	
	    }			
	    tmp ++;			
	    scrn -= buffer_width - BUFF_WIDTH;
	}
    } else {
	screen.changeTo(CHANGE_NEXT);
	return 1;
    }
    return 0;
}


int screen_4kal() {
    if( BUFF_WIDTH/2 <= BUFF_HEIGHT) {
	unsigned char * tmp, * scrn;
	int x,y;
	
	tmp = active_buffer;
	scrn = display_mem;
	/* upper half */
	for(y=BUFF_HEIGHT/2; y != 0; y--) {
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;			
		scrn ++;
		tmp += BUFF_WIDTH;		
	    }				
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;		
		scrn ++;
		tmp -= BUFF_WIDTH;		
	    }				
	    tmp ++;				
	    scrn += buffer_width - BUFF_WIDTH;
	}

	tmp = active_buffer;
	scrn = display_mem + buffer_width * (BUFF_HEIGHT-1) + BUFF_WIDTH-1;
	/* lower half */
	for(y=BUFF_HEIGHT/2; y != 0; y--) {
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;		
		scrn --;
		tmp += BUFF_WIDTH;		
	    }				
	    for(x=BUFF_WIDTH/2; x != 0; x--) {
		*scrn = *tmp;		
		scrn --;
		tmp -= BUFF_WIDTH;		
	    }				
	    tmp ++;				
	    scrn -= buffer_width - BUFF_WIDTH;
	}
    } else {
	screen.changeTo(CHANGE_NEXT);
	return 1;
    }
    return 0;
}



/*
 * 3-D display functions
 */
static const float SC = 1 << 16;
static xy height_offset[256];
static xy s1, s2, p;
static float rot[3] = {0,0,0};
static float P[4][3] = { {-1,-1,0}, {1,-1,0}, {1,1,0}, {0,0,1} };

/*
 * rotate vektor p around x/y/z axis (by rot[i])
 * result in vector r
 */
void rotate(float p[3], float rot[3], float r[3]) {
    float t1[3], t2[3];
    /* first around x axis */
    t1[0] = p[0];
    t1[1] = p[1] * cos(rot[0]) + p[2] * sin(rot[0]);
    t1[2] = -p[1] * sin(rot[0]) + p[2] * cos(rot[0]);

    /* around y axis */
    t2[0] = t1[0] * cos(rot[1]) + t1[2] * sin(rot[1]);
    t2[1] = t1[1];
    t2[2] = -t1[0] * sin(rot[1]) + t1[2] * cos(rot[1]);

    /* around z axis */
    r[0] = t2[0] * cos(rot[2]) + t2[1] * sin(rot[2]);
    r[1] = -t2[0] * sin(rot[2]) + t2[1] * cos(rot[2]);
    r[2] = t2[2];
}
inline float sqr(float a) {
    return a*a;
}

/*
 * preparations for a 3D display
 */
int prepare_3d(int maxZ) {
    int i;
    float x,y,z,l;
    float ro[4][3];
    double scale;

    /* calculate the maximal size of the "sheet" */
    x = sqr(BUFF_WIDTH/2);
    y = sqr(BUFF_HEIGHT/2);
    z = sqr((float)(BUFF_WIDTH)/640.0 * (float)maxZ);
    l = sqrt(x + y + z);

    /* make sure, the image always fits into the display region */
    scale = (float)min(BUFF_WIDTH, BUFF_HEIGHT) / l;
    
    /* rotate a little bit */
    /* TODO: use fps, some value from the sound to set speed,
       external value (joystick) to set rotation */
    rot[0] += 0.006;
    rot[1] += 0.010;
    rot[2] += 0.004;
    for(i=0; i < 4; i++)
	rotate(P[i], rot, ro[i]);
	
    for(i=0; i < 256; i++) {
	height_offset[i].x = int( ro[3][0] * i * double(BUFF_WIDTH) / 640.0 * SC);
	height_offset[i].y = int( ro[3][1] * i * double(BUFF_WIDTH) / 640.0 * SC);
    }

    /* calculate the step sizes */
    s1.x = int( (ro[1][0] - ro[0][0]) * scale * SC/2.0);
    s1.y = int( (ro[1][1] - ro[0][1]) * scale * SC/2.0);

    s2.x = int( (ro[2][0] - ro[1][0]) * scale * SC/2.0);
    s2.y = int( (ro[2][1] - ro[1][1]) * scale * SC/2.0);

    /* starting point */
    p.x = -BUFF_WIDTH/2 * s1.x -BUFF_HEIGHT/2 * s2.x;
    p.y = -BUFF_WIDTH/2 * s1.y -BUFF_HEIGHT/2 * s2.y;

    /* clear screen */
    unsigned char * clr = display_mem;
    for(i=2*BUFF_HEIGHT; i != 0; i--) {
	bzero(clr, 2*BUFF_WIDTH);
	clr += buffer_width;
    }

    return 0;
}

/*
 * hfield: buffer_value determines height
 */
int screen_hfield() {
    unsigned char * src = active_buffer;
    unsigned char * dst = display_mem + BUFF_WIDTH + buffer_width*BUFF_HEIGHT;
    int i,j;
    
    if( (2*BUFF_WIDTH > disp_size.x) || (2*BUFF_HEIGHT > disp_size.y) ) {
	screen.changeTo(CHANGE_NEXT);
	return 1;
    }

    prepare_3d(256);

    for(i=BUFF_HEIGHT; i != 0; i--) {
	xy t = p;			
	for(j=BUFF_WIDTH; j != 0; j--) {	
	    unsigned char * d = dst +	
		((p.y + height_offset[*src].y) >> 16) * buffer_width +	
		((p.x + height_offset[*src].x) >> 16);	
	    *d = *src | 128;			
	    src++;			
	    p.x += s1.x; 
	    p.y += s1.y;	
	}				
	p.x = t.x + s2.x;	
	p.y = t.y + s2.y;		
    }

    return 0;
}

/*
 * bent: height is a sine-wave along x axis
 */
int screen_bent() {
    unsigned char * src = active_buffer;
    unsigned char * dst = display_mem + BUFF_WIDTH + buffer_width*BUFF_HEIGHT;
    int i,j;
    static int height[MAX_BUFF_WIDTH];
    static float t = 0;
    static float h = 0.0;
    static float stp = 0.0;

    if( (2*BUFF_WIDTH > disp_size.x) || (2*BUFF_HEIGHT > disp_size.y) ) {
	screen.changeTo(CHANGE_NEXT);
	return 1;
    }

    prepare_3d(256);

    h = h * 0.95 + ((2.0*sound_amplitude) <? 120.0) * 0.05;

    for(i=BUFF_WIDTH; i != 0; i--) {
	height[i] = (int)(h * sin(t) *			\
			  sin((double)(i)/(double)BUFF_WIDTH*3.0*M_PI)) + 128;
    }
    stp = stp * 0.95 + ((fps*2.0) >?  1) * 0.05;
    t += 4.0/stp;

    for(i=BUFF_HEIGHT; i != 0; i--) {
	xy t = p;	
	for(j=BUFF_WIDTH; j != 0; j--) {
	    unsigned char * d = dst +	
		((p.y + height_offset[height[j]].y) >> 16) * buffer_width +
		((p.x + height_offset[height[j]].x) >> 16);
	    *d = *src | 128;			
	    src++;			
	    p.x += s1.x; 
	    p.y += s1.y;	
	}				
	p.x = t.x + s2.x;		
	p.y = t.y + s2.y;		
    }
    
    return 0;
}

/*
 * plate: height contant to 0
 */
int screen_plate() {
    unsigned char * src = active_buffer;
    unsigned char * dst = display_mem + BUFF_WIDTH + buffer_width*BUFF_HEIGHT;
    int i,j;

    if( (2*BUFF_WIDTH > disp_size.x) || (2*BUFF_HEIGHT > disp_size.y) ) {
	screen.changeTo(CHANGE_NEXT);
	return 1;
    }

    prepare_3d(1);

    for(i=BUFF_HEIGHT; i != 0; i--) {	
	xy t = p;			
	for(j=BUFF_WIDTH; j != 0; j--) {	
	    unsigned char * d = dst + (p.y >> 16) * buffer_width + (p.x >> 16);
	    *d = *src | 128;
	    src++;			
	    p.x += s1.x; 
	    p.y += s1.y;	
	}				
	p.x = t.x + s2.x;		
	p.y = t.y + s2.y;		
    }

    return 0;
}

/*
 * Map the buffer on a cylinder (around x-axis), and roll it.
 * Buffer is mapped 2 times on the cylinder, so that no discontinuities arise
 *
 * other possible things with this idea: waves, ...
 */
int screen_roll() {
    int i;
    static double theta = 0;		/* rotation angle */

    for(i=0; i < BUFF_HEIGHT; i++) {
	int p = i - BUFF_HEIGHT/2;		
	double phi = acos( (double)p / (double)(BUFF_HEIGHT/2));
	int b = (int)((theta + phi) / M_PI * (double)BUFF_HEIGHT);
	b %= 2*BUFF_HEIGHT; 
	if (b < 0)		b += 2*BUFF_HEIGHT; 
	if (b >= BUFF_HEIGHT)	b = 2*BUFF_HEIGHT - b - 1;

	perm_lines[i] = PERM_ADDR(b);
    }
    theta += M_PI / 40.0;		/* roate a little bit */
    if (theta > 2.0*M_PI) 
	theta -= 2.0*M_PI;

    screen_perm();

    return 0;
}


/*
 * this is not looking good
 */
#define ZICK_SMOOTH	10
static int zicks[MAX_BUFF_HEIGHT];
int screen_zick() {
    int i;
    unsigned char * scrn = display_mem;
    unsigned char * src = active_buffer;
    static int first = 1;

    static int d = 0; 

    if(first) {
	for(i=0; i < BUFF_HEIGHT+1; i++)
	    zicks[i] = 0;
    }

    for(i=BUFF_HEIGHT; i != 0; i--) {
	zicks[i] = ((sound_data[i][0]) + ZICK_SMOOTH*zicks[i]) / (ZICK_SMOOTH+1);
        d = (d + zicks[i])/2;
        if(d == 0) {		
	    memcpy(scrn, src, BUFF_WIDTH);
	} else if(d > 0) {			
	    memset(scrn, 0, d);			
	    memcpy(scrn+d, src, BUFF_WIDTH-d);	
        } else {				
	    memcpy(scrn, src-d, BUFF_WIDTH+d);	
	    memset(scrn+disp_one_size.x+d, 0, -d);
	}			
        src += BUFF_WIDTH;	
	scrn += buffer_width;	
    }

    return 0;
}


