/*
    CTHUGHA-L							display.h
	
    Funktions to access the Screen.
	- general stuff
	- screen-functions
	- palette-functions
	- text-functions
*/

#ifndef __DISPLAY_H__
#define __DISPLAY_H__

#include <time.h>

#include "Option.h"

/* 
 *  General stuff 
 */
int init_ncurses();						
int init_display();				/* initialize VGA */
int init_graph_mode();				/* activate graphics-mode */
int exit_graph_mode();				/* return to text-mode */
int exit_display();				/* clean-up */
int exit_ncurses();

int resize_display(int new_width, int new_height);

/*
 * variables, that have to do with the screen size and tiles
 *
 * disp_size: size of the screen in svga-mode
 *            size of the allocated imaged in xwin-mode
 *
 * disp_one_size: size of one drawn buffer (as drawn by a screen function)
 * disp_size: total size of all the drawn buffers
 *
 */
extern xy disp_size;				/* size of drawing area */
extern xy disp_one_size;
extern xy draw_size;				/* size of the drawn image */

extern int bytes_per_line;
extern unsigned char * active_buffer;		/* buffer next on screen */
extern unsigned char * passive_buffer;		/* buffer current on screen */

extern int display_mode;			/* graphic mode */
extern int display_direct;			/* draw directly to screen */
extern unsigned char * display_mem;
extern int display_frames;			/* counter for speed-tests */
extern time_t display_start;
extern int display_mit_shm;			/* use MIT-SHM if possible */
extern int display_on_root;			/* display on root window */
extern int bpp;					/* bits per pixel */
extern int bypp;				/* bytes per pixel */
extern unsigned long bitmap_colors0[256];	/* "compiled" palette */
extern unsigned long bitmap_colors1[256];	
extern unsigned long bitmap_colors2[256];	
extern unsigned long bitmap_colors3[256];	
extern int rev_byte_order;
enum draw_mode_t {
    DM_direct,
    DM_tmp_mapped,
    DM_mapped1,
    DM_mapped2,
    DM_mapped3,
    DM_mapped4
};
extern enum draw_mode_t draw_mode;		/* how drawing is done */

extern int ncurses_use;

/* 
 *  Stuff about screen-functions
 */
int display();					/* bring the changes to scr. */

extern char screen_first[256];			/* Start with this func. */
extern int display_syncwait;			/* Wait for sync on display */

/* 
 *  Stuff about palettes
 */
typedef unsigned char Palette[256][3];		/* one Palette: 256 entries, 
						   each 3 bytes */
class PaletteEntry : public OptionListEntry {
public:
    Palette pal;

    PaletteEntry(const char * name, const char * desc) :
	OptionListEntry(name, desc) {}
    PaletteEntry(FILE * file, const char * name);

    friend OptionListEntry * read_paleette(FILE *, const char *, const char *);
};

int load_palettes();				/* initializiation */
int init_palettes();
int exit_palettes();
extern char palette_first[256];			/* Start with this palette */
int update_palette();
extern int display_internal_pal;		/* disable internal pal. */
extern int display_external_pal;		/*  disable external pal. */
void cth_setpalette(Palette pal, int immed);
void smooth_setpalette();
extern int display_private_cmap;		/* use a private colormap */
void set_rgb_palette();

/* 
 *  Stuff about text-display
 */
#define TEXT_COLOR_NORMAL		0
#define TEXT_COLOR_ERROR		1
#define TEXT_COLOR_HIGHLIGHT		2

void display_text_erase();
void display_text_refresh();
void display_print(const char * text,	/* what text to display */
		   double y,		/* line number */
		   int justifcation,	/* 'l', 'c', 'r' */
		   int color);		/* colorindex to use 255, 254, .. */
extern xy text_size;

extern int display_text_time;

extern int display_text_on_screen;		/* text is currently on scrn */

extern int display_text_color[][3];
extern int display_text_colors;

/* 
 *  Stuff for PCX
 */
typedef unsigned char * Pcx;			/* Memory for loaded picture */
int init_pcx();					/* initalize pcx */
int show_pcx(int upp_left, int correct_color);
void smooth_setpalette(void);
extern int display_use_pcx;			/* allow pcx-usage */
extern char pcx_first[256];
extern int pcx_count;				/* counter for pcx-display */
extern int * pcx_palettes;			/* index to corresp. palette */

extern char display_prt_file[];			/* filename used by PrtScrn */
int save_buffer();
int save_display();

#endif


