#include "cthugha.h"
#include "display.h"
#include "interface.h"
#include "disp-sys.h"
#include "action.h"
#include "imath.h"

#include "initial_palettes.cc"

int change_palette_imm(int);

OptionList palette("palette", NULL, NULL, 0);

char palette_first[256] = "";			/* Start with this palette */

int display_internal_pal = 1;			/* use internal pal. */
int display_external_pal = 1;			/* use external pal. */

unsigned long bitmap_colors0[256];		/* "compiled" palette */
unsigned long bitmap_colors1[256];		/* "compiled" palette */
unsigned long bitmap_colors2[256];		/* "compiled" palette */
unsigned long bitmap_colors3[256];		/* "compiled" palette */

OptionListEntry * read_palette(FILE * file, const char * name, const char * dir);
static const char * palette_path[] = {
    "./",
    "./map/",
    CTH_LIBDIR "/map/",
    ""	
};

int colormapped=1;				/* 0 .. True/Direct color
						   1 .. All 256 color cells
						   2 .. Only some cells */

Palette rgb_pal;				/* palette for RGB and 256 colors */


/* when using only some colorcells (e.g. when drawing on the root window)
   update to the palette are done immediately without smoothing.
   */

int load_palettes() {
    int i,j,k,l;
	
    /* Builtin palettes */
    if ( display_internal_pal ) {
	char str[128];
	PaletteEntry * new_pal;
	Palette * pal;

	printfv(2,"  preparing internal palettes...\n");

	for(i=0; (unsigned int)i < sizeof(initial_palettes)/sizeof(Palette); i++) {
	    sprintf(str, "Internal_%d", i);
	    new_pal = new PaletteEntry(str, "");
	    memcpy(new_pal->pal, initial_palettes[i], sizeof(Palette));
	    palette.add(new_pal, new OptionOnOff("", 1)); 
	}

	/* create one general palette */
	sprintf(str, "Internal_%d", palette.getNEntries());
	new_pal = new PaletteEntry(str, "");
	pal = &(new_pal->pal);
	for(i=0; i < 64; i++) {
	    (*pal)[i][0]	= i << 2;
	    (*pal)[i][1]	= 0;
	    (*pal)[i][2]	= 0;

	    (*pal)[i+64][0] 	= 0;
	    (*pal)[i+64][1]	= i << 2;
	    (*pal)[i+64][2] 	= 0;

	    (*pal)[i+128][0] 	= 0;
	    (*pal)[i+128][1]	= 0;
	    (*pal)[i+128][2] 	= i << 2;

	    (*pal)[i+192][0] 	= i << 2;
	    (*pal)[i+192][1]	= i << 2;
	    (*pal)[i+192][2] 	= i << 2;
	}
	palette.add(new_pal, new OptionOnOff("", 1));
    }

    /* create the RGB pal for 256 colors */
    for(l=0, i=0; i < 8; i++)
	for(j=0; j < 8; j++)
	    for(k=0; k < 4; k++, l++) {
		rgb_pal[l][0] = i << 5;
		rgb_pal[l][1] = j << 5;
		rgb_pal[l][2] = k << 6;
	    }
    
    /* read palettes from file */
    if ( display_external_pal) {
	printfv(2,"  loading external palettes...\n");
	palette.load(palette_path, "/map/", ".map", read_palette);
	printfv(2,"  number of loaded palettes: %d\n", palette.getNEntries());
    }

    
    /* brighten up palettes, that are a bit dark */
    for(i=0; i < palette.getNEntries(); i++) {
	Palette * pal = &(((PaletteEntry*)palette[i])->pal);
	int m = 0;
	double P = 0;

	for(l=0; l < 256; l++) {
	    m = max(m, (*pal)[l][0] + (*pal)[l][1] + (*pal)[l][2]);
	}
	if( (m > 0) && (m < 3*255) ) {
	    P = double(3*255)/double(m);
	    printfv(10, "    brightening palette %d (%s). Faktor: %0.3f\n", 
		    i, palette[i]->Name(), P);

	    for(l=0; l < 256; l++) {
		(*pal)[l][0] = min((int) ( (double)(*pal)[l][0] * P), 255);
		(*pal)[l][1] = min((int) ( (double)(*pal)[l][1] * P), 255);
		(*pal)[l][2] = min((int) ( (double)(*pal)[l][2] * P), 255);
	    }
	}
    }

    return 0;
}

/*
 * store the palette with 8 bytes per color
 */
OptionListEntry* read_palette(FILE * file, const char * name, const char * /*dir*/) {
    char dummy[256];
    int i, j, r,g,b;
    Palette * pal;
    PaletteEntry * new_pal = new PaletteEntry(name, "");

    pal = &(new_pal->pal);

    for(i=0; i < 256; i++) { 
	if ( fscanf( file, "%d %d %d", &r, &g, &b) < 3) {
	    printfv(3,"\n    Can't read at line: %d (%s)", i,name);
	    if ( i == 0) {				/* nothing read */
		printfv(3," ... skipping file");
		delete new_pal;
		return NULL;
	    }
	    printfv(3," ... filling with black");
	    for(; i < 256; i++)				/* fill with black */
		for(j=0;j<3;j++)
		    (*pal)[i][j] = 0;
	    break;
	}
	fgets( dummy, 255, file);
	(*pal)[i][0] = r;
	(*pal)[i][1] = g;
	(*pal)[i][2] = b;
    }

    return new_pal;
}

/*
 * Set the first palette
 *   - allocate palette
 */
int init_palettes() {
    if( init_palettes_sys() )
	return 1;

    if(int(rgb))
	set_rgb_palette();

    update_palette(); 
    return 0;
}

int exit_palettes() {
    exit_palettes_sys();
    return 0;
}

/*
 * Functions to change/selct the palette
 *
 * changes (smooth_palette) by "Stanislav V. Voronyi" <stas@use.kharkov.ua>
 */
Palette cur_palette;
static Palette new_palette;
static int pal_changed = 0;
static int pal_text = 0;	/* a "text-palette" is on screen */


void cth_setpalette(Palette pal, int immed) {
    static int cur_empty = 1;

    memcpy( new_palette, pal, sizeof(Palette) );	/* save new palette */

    if ( display_text_on_screen) {			/* text-mode */
	int i,j;

	/* create a dimmed palette */

	if(colormapped ) {		/* use only colors 0 - 127, save rest for text */
	    for(i=0; i < 128; i++)
		for(j=0; j < 3; j++)
		    new_palette[i][j]  = (new_palette[i*2][j] + new_palette[i*2+1][j]) >> 2;

	    /* text colors */
	    for(i=0; i < display_text_colors; i++)
		for(j=0; j < 3; j++) {
		    cur_palette[255-i][j] = new_palette[255-i][j] = display_text_color[i][j];
		}
	} else {			/* here we can use all colors */
	    for(i=0; i < 256; i++)
		for(j=0; j < 3; j++)
		    new_palette[i][j]  >>= 1;
	}

	/* use only some of the colors, when text is on screen */
	if( draw_mode == DM_direct) {
	    draw_mode = DM_tmp_mapped;

	    for(i=0; i < 256; i++) {
		bitmap_colors0[i] = i / 2;
		bitmap_colors1[i] = bitmap_colors0[i] << 8;
		bitmap_colors2[i] = bitmap_colors1[i] << 8;
		bitmap_colors3[i] = bitmap_colors2[i] << 8;
	    }
	}
	    
	pal_text = 1;
    } else {
	pal_text = 0;
	if(draw_mode == DM_tmp_mapped)		/* switch back to direct draw mode */
	    draw_mode = DM_direct;
    }

    pal_changed = 1;				/* palette entries changed */

    if( cur_empty || immed || (colormapped == 2) ) {
 	memcpy( cur_palette, new_palette, sizeof(Palette) );
	cur_empty=0;

	smooth_setpalette();			/* really set the palette */
    }
}

void set_rgb_palette() {
    Palette tmp;

    memcpy(tmp, cur_palette, sizeof(Palette));
    memcpy(cur_palette, rgb_pal, sizeof(Palette));
    set_palette_sys();
    memcpy(cur_palette, tmp, sizeof(Palette));

}

/*
 * brings the current palette nearer to the desired palette (new_palette)
 * smoothing one palette to the next needs at most 1 second.
 */
void smooth_setpalette(void) {
    int i,j;
    int max_chg;

    // check if the palette changed
    static int current_pal = -1;

    if( (current_pal != palette.currentN())  ||		// palette changed
	(display_text_on_screen && !pal_text) ||	// text came on screen
	(!display_text_on_screen && pal_text) ) {	// text left screen

	PaletteEntry * p = (PaletteEntry*)palette.current();
	if(p == NULL)
	    return;
	
	current_pal = palette.currentN();

	cth_setpalette(p->pal, 0);
    }

    if (! pal_changed )				/* no update needed */
	return;

    max_chg = (fps > 0) ? max(int(255.0 / fps), 1) : 4;

    /* count if anything changed */
    pal_changed = 256*3;
    for(i=0; i < 256; i++)
	for(j=0; j < 3; j++) {
	    int d = new_palette[i][j] - cur_palette[i][j];
	    d = min( max(d, -max_chg), max_chg);	/* d between -max_chg and +max_chg */
	    if( d != 0)
		cur_palette[i][j] += d;
	    else
		pal_changed --;
	}

    /* really set the palette now, only if not running in RGB mode */
    if(! int(rgb))
	set_palette_sys();
}

/*
 * set a new palette
 */
int update_palette() {

    PaletteEntry * p = (PaletteEntry*)palette.current();
    if(p == NULL)
	return 0;

    cth_setpalette(p->pal, 1);

    return 0;
}












