#include "cthugha.h"
#include "interface.h"
#include "keys.h"
#include "sound.h"
#include "imath.h"

#include <ctype.h>
#include <signal.h>

extern int nInterfaces;	// this is at the bottom of this file
int interfaceFktCurrent = -1;

int interfaceToggle = 1;

int setInterface(Interface * interface) {
    if(interface != NULL) {
	currentInterface = interface;

	// check if this can also be reached by a function key
	interfaceFktCurrent = -1;
	for(int i=0; i < nInterfaces; i++)
	    if(*interfaceFkt[i] == interface) 
		interfaceFktCurrent = i;

	return 0;
    }
    return 1;
}
int setInterface(int nr) {
    if( (nr < nInterfaces) && (nr >= 0) ) {
	currentInterface = *(interfaceFkt[nr]);
	interfaceFktCurrent = nr;
	if(currentInterface == NULL)
	    nextInterface();
	return 0;
    }
    return 1;
}

int nextInterface() {
    if( (interfaceFktCurrent >= 0) && (interfaceFktCurrent < nInterfaces)) {
	
	do {
	    interfaceFktCurrent = mod(interfaceFktCurrent+1, nInterfaces);
	} while( *(interfaceFkt[interfaceFktCurrent]) == NULL);
	currentInterface = *(interfaceFkt[interfaceFktCurrent]);

	return 0;
    }
    return 1;
}
int prevInterface() {
    if( (interfaceFktCurrent >= 0) && (interfaceFktCurrent < nInterfaces)) {

	do {
	    interfaceFktCurrent = mod(interfaceFktCurrent-1, nInterfaces);
	} while( *(interfaceFkt[interfaceFktCurrent]) == NULL);
	currentInterface = *(interfaceFkt[interfaceFktCurrent]);

	return 0;
    }
    return 1;
}


ErrorMessages errors;


//
// standard interface
//

// default key handler
int Interface::do_key(int key) {

    int new_sel = -1;
    int jump = 0, dir = 1;

    switch(key) {
    case CK_HOME:
	new_sel = 0;	dir = 1;
	break;
    case CK_PGUP:
	jump = -10;	dir = -1;
	break;
    case CK_UP:
	jump = -1;	dir = -1;
	break;
	
    case CK_END:
	new_sel = nElements - 1; dir = -1;
	break;
    case CK_PGDN:
	jump = 10;	dir = 1;
	break;
    case CK_DOWN:
	jump = 1;	dir = 1;
	break;

    case 'i': case 'I':
	if( (currentInterface == interfaceInfo) && interfaceToggle)
	    setInterface(interfaceMain);
	else
	    setInterface(interfaceInfo);
	return 0;
 
   case 'o': case 'O':
	if( (currentInterface == interfaceOptions) && interfaceToggle)
	    setInterface(interfaceMain);
	else
	    setInterface(interfaceOptions);
	return 0;

    case '?': case '':		//  is the on the same key as ? on a german keyboard
	key = CK_FKT(1);		

    case CK_FKT(1):	case CK_FKT(2):
    case CK_FKT(3):	case CK_FKT(4):
    case CK_FKT(5):	case CK_FKT(6):
    case CK_FKT(7):	case CK_FKT(8):
    case CK_FKT(9):	case CK_FKT(10):
    case CK_FKT(11):	case CK_FKT(12): 
	key = key - CK_FKT(1);		// convert keycode to integer

	if(key >= nInterfaces)
	    return key;
	
	// keys act as toggle
	if( (*interfaceFkt[key] == currentInterface) && interfaceToggle) {
	    setInterface(interfaceMain);
	} else {
	    setInterface(key);
	}
	return 0;
	
    case 'q': case 'Q': case CK_ESC:
	setInterface(interfaceMain);
	return 0;
	
    default:
	if( sel < nElements)
	    return elements[sel]->do_key(key);
    }

    // update selection line
    if(jump && (nElements > 1)) {
	jump = jump % nElements;
	new_sel = (sel + nElements + jump) % nElements;
    }
    if( (new_sel >= 0) && (nElements > 1)) {
	while( !elements[new_sel]->selectable && (sel != new_sel) )
	    new_sel = (new_sel + nElements + dir) % nElements;
	
	if(sel == new_sel)
	    return 0;
	
	elements[sel]->selected = 0;
	sel = new_sel;
	elements[new_sel]->selected = 1;
	return 0;
    }

    return key;
}


// default display handler
void Interface::display() {
    for(int i=0; i < nElements; i++)
	elements[i]->display(i);
}

// default runner
void Interface::run() {

    // handle keys
    int key;
    while( (key = getkey()) != CK_NONE) {
	this->do_key(key);
    }

    if(cthugha_pause) {
	/* this is here to be sure not to interrupt a graphics operation. */
	cthugha_pause = 0;

	exit_graph_mode();			/* turn off graphics */
	exit_sound();				/* and sound */
	
	raise(SIGTSTP);				/* default action */
    }
}













void InterfaceElement::display(int y) {
    display_print(text, y, 'l', selected ? c2 : c1);
}


int InterfaceElement::do_key(int key) {
    return key;
}







int InterfaceElementTitle::do_key(int key) {
	
    switch(key) {
    case '+': case '>':
    case CK_RIGHT:
	if( nextInterface() )
	    return key;
	return 0;
    case '-': case '<':
    case CK_LEFT:
	if( prevInterface() )
	    return key;
	return 0;
    }
    return key;
};


void InterfaceElementOption::display(int line) {
    char str[128];
    char fmt[128];

    // make format and include the > <
    sprintf(fmt, "%c%s%c", 
	    selected ? '>' : ' ',
	    text,
	    selected ? '<' : ' ');

    // insert the text of the option
    sprintf(str, fmt, option->text());

    // and bring it to the screen
    display_print(str, line, 'l', selected ? c2 : c1);
}

int InterfaceElementOption::do_key(int key) {

    switch(key) {
    case '+':
	option->changeBy(inc1);
	return 0;
    case '-':
	option->changeBy(-inc1);
	return 0;
    case CK_RIGHT:
	option->changeBy(inc2);
	return 0;
    case CK_LEFT:
	option->changeBy(-inc2);
	return 0;
    case '.': case ',':
	option->changeBy(inc3);
	return 0;
    case '0':
	option->changeBy(-inc3);
	return 0;
    }
    return key;
}



void ErrorMessages::addMessage(const char * text) {
    if(nMsgs == 128) {
	printfe("too many errors: %s\n", text);
	return;
    }
    strncpy(msgs[nMsgs], text, 128);
    on_screen[nMsgs] = gettime();

    nMsgs ++;
}
void ErrorMessages::display() {
    
    // bring messages to screen
    for(int i = 0; i < nMsgs; i ++) {
	display_print( msgs[i], - (nMsgs-i), 'r', TEXT_COLOR_ERROR);
    }

    // remove old messages
    while( (nMsgs > 0) && ((gettime() - on_screen[0]) > display_text_time) ) {
	for(int i = 1; i < nMsgs; i ++) {
	    strncpy(msgs[i-1], msgs[i], 128);
	    on_screen[i-1] = on_screen[i];
	}
	nMsgs --;
    }
}

    




Interface ** interfaceFkt[] = {
    &interfaceHelp,
    &interfaceScreen,
    &interfaceFlame,
    &interfaceWave,
    &interfacePalette,
    &interfaceTranslate,
    &interfacePCX,
    &interfaceObject,
    &interfaceCD,
    &interfaceSound,
    &interfaceMixer,
    &interfaceServer
};
int nInterfaces = sizeof(interfaceFkt)/sizeof(Interface*);

Interface * currentInterface = interfaceMain;

