// -*-c++-*-

#ifndef __INTERFACE_H
#define __INTERFACE_H

#include "display.h"

class InterfaceElement {
protected:
    const char * text;
    int c1, c2;
    int selectable;
    int selected;
public:
    InterfaceElement(const char * t, int c1, int c2, int sel) : 
	text(t), c1(c1), c2(c2), selectable(sel), selected(0) {
    }

    virtual void display(int line);
    virtual int do_key(int key);

    friend class Interface;
    friend class InterfaceList;
};

//
// a line of text
//
class InterfaceElementText : public InterfaceElement {
public:
    InterfaceElementText(const char * t) : 
	InterfaceElement(t, TEXT_COLOR_NORMAL, TEXT_COLOR_NORMAL, 0) {}
};

//
// the title line
//
class InterfaceElementTitle : public InterfaceElement {
public:
    InterfaceElementTitle(const char * t) : 
	InterfaceElement(t, TEXT_COLOR_NORMAL, TEXT_COLOR_HIGHLIGHT, 1) {}

    int do_key(int key);
};

//
// an option line
//
class InterfaceElementOption : public InterfaceElement {
    Option * option;
    int inc1, inc2, inc3;
public:
    InterfaceElementOption(const char * fmt, Option * opt, int i1=1, int i2=1, int i3=1) : 
	InterfaceElement(fmt, TEXT_COLOR_NORMAL, TEXT_COLOR_HIGHLIGHT, 1),
	option(opt), inc1(i1), inc2(i2), inc3(i3) {}

    void display(int line);
    int do_key(int key);
};




//
// The user interface
//
class Interface {
protected:
    InterfaceElement ** elements;
    int nElements;

    int sel;

    virtual int do_key(int key);
public:
    Interface(InterfaceElement ** el, int nEl) : 
	elements(el), nElements(nEl), sel(0) {
	if(nEl > 0)
	    elements[0]->selected = 1;
    }
    
    virtual void run();
    virtual void display();

    friend class InterfaceElement;
};

//
// user Interface for Lists
//
class InterfaceList : public Interface {
    static int size;
    int pos;
    OptionList * optList;
    int nFixed;
public:
    InterfaceList(InterfaceElement ** fixed, int nEl, OptionList * ol);
    void display();
    int do_key(int key);
};

//
// the main user interface
//
class InterfaceMain : public Interface {
    char * silenceMsg;
    int silenceTime;
    double silenceLine;
public:
    InterfaceMain();
    InterfaceMain(InterfaceElement ** el, int nEl) :
	Interface(el, nEl), silenceMsg(NULL) {}
    int do_key(int key);
    void display();
    void msg(char * msg);
};

//
// Help text
//
class InterfaceHelp : public InterfaceMain {
public:
    InterfaceHelp(InterfaceElement ** el, int nEl) :
	InterfaceMain(el, nEl) {}
    void display() { Interface::display(); }
};

//
// Server client lists
//
class InterfaceServer : public Interface {
public:
    InterfaceServer(InterfaceElement ** el, int nEl) :
	Interface(el, nEl) {}
    void display();
};


extern Interface * currentInterface;
int setInterface(Interface * interface);
int setInterface(int nr);
int nextInterface();
int prevInterface();

extern Interface * interfaceMain;
extern Interface * interfaceHelp;
extern Interface * interfaceInfo;
extern Interface * interfaceCredits;		// in InterfaceCredits.cc
extern Interface * interfaceOptions;
extern Interface * interfaceCD;
extern Interface * interfaceSound;
extern Interface * interfaceMixer;
extern Interface * interfaceScreen;
extern Interface * interfaceFlame;
extern Interface * interfaceWave;
extern Interface * interfacePalette;
extern Interface * interfaceTranslate;
extern Interface * interfacePCX;
extern Interface * interfaceObject;
extern Interface * interfaceServer;

extern Interface ** interfaceFkt[];


//
// an error message
//
class ErrorMessages {
    char msgs[128][128];
    int on_screen[128];
    int nMsgs;
public:
    ErrorMessages() : nMsgs(0) {}

    void addMessage(const char * text);
    void display();
};


extern ErrorMessages errors;

extern interfaceToggle;

#endif








