#include "cthugha.h"
#include "sound.h"
#include "display.h"
#include "cd_player.h"
#include "information.h"
#include "interface.h"
#include "server.h"
#include "imath.h"
#include "cth_buffer.h"

char2 * sound_data;			/* Sound-Data in 8 bit signed, stereo */
char2 * sound_data_unproc;		/* Sound-Data in 8 bit signed, stereo, unprocessed */
int sound_bsize = 0;			/* size of sound-buffer (samples) 
					   always a power of 2 */


SoundDevice * soundDevice = NULL;
OptionInt sound("sound-device-number", SDN_DSPIn, SDN_Max);

OptionOnOff sound_fork("sound-fork", 0);	// read in an extra process

OptionSound sound_sample_rate("sound-sample-rate", 44000);
OptionChannels sound_channels("sound-channels", 2);
OptionSound sound_sample_size("sound-sample-size", 8);
OptionSound sound_method("sound-method", 0, 3);			// only for DSP
OptionSound sound_fragments("sound-fragments", 16);		// only for DSP
OptionSound sound_fragment_size("sound-fragment-size", 0);	// only for DSP
OptionOnOff sound_sync("sound-sync", 0);			// only for DSP


void OptionSound::changeBy(int by) {
    OptionInt::changeBy(by);

    if(soundDevice)
	soundDevice->update();
}	


//
// the interface
//
static InterfaceElement * elementsSound[] = {
    new InterfaceElementTitle("Sound Interface"),
    new InterfaceElementText("----------------------------------------"),
    new InterfaceElementOption("Sample rate       : %10s Hz", &sound_sample_rate, 50, 100, 500),
    new InterfaceElementOption("Channels          : %10s", &sound_channels),
    new InterfaceElementOption("Sample Size       : %10s bit", &sound_sample_size, 8, 8, 8),
    new InterfaceElementOption("DSP Method        : %10s", &sound_method),
    new InterfaceElementOption("DSP Fragments     : %10s", &sound_fragments),
    new InterfaceElementOption("DSP Fragment Size : %10s", &sound_fragment_size),
    new InterfaceElementOption("DSP Sound sync    : %10s", &sound_sync),
    new InterfaceElementOption("File buffer size  : %10s k", &soundBuffer),
    new InterfaceElementOption("File play loop    : %10s", &soundLoop)
};
Interface * interfaceSound = new Interface(elementsSound, 
					 sizeof(elementsSound)/sizeof(InterfaceElement*) );



int newSoundDevice(SoundDeviceNr sDev) {

    // create the right device
    switch(sDev) {
    case SDN_DSPIn:
	soundDevice = new SoundDeviceDSPIn();
	break;
    case SDN_Net:
	soundDevice = new SoundDeviceNet();
	break;
    case SDN_Random:
	soundDevice = new SoundDeviceRandom();
	break;
    case SDN_File:
	soundDevice = new SoundDeviceFile(0);
	break;
    case SDN_FileLoud:
	soundDevice = new SoundDeviceFile(1);
	break;
    case SDN_Exec:
	soundDevice = new SoundDeviceExec(0);
	break;
    case SDN_ExecLoud:
	soundDevice = new SoundDeviceExec(1);
	break;
    case SDN_Fork:
	soundDevice = new SoundDeviceFork();
	break;
    default:
	printfe("Illegal SoundDeviceNr %d.\n", sDev);
	return 1;
    }

    // check for errors
    if( soundDevice->err() != 0) {
	delete soundDevice;
	soundDevice = NULL;
	return 1;
    }
    return 0;
}


/* 
 *  Initialize the Sound-Interface
 */
int init_sound() {

    sound_bsize =  1 << ilog2(max(BUFF_WIDTH, BUFF_HEIGHT));

    /* check, if already allocated */
    if(soundDevice != NULL)
	return 0;

    sound_data_unproc = new char2[sound_bsize];
    sound_data = new char2[sound_bsize];

    return int(sound_fork) ? newSoundDevice(SDN_Fork) : newSoundDevice(int(sound));
}


/*
 * Clean up the sound-interface
 */
int exit_sound() {
    delete soundDevice;
    soundDevice = NULL;
    return 0;
}


/*****************************************************************************/

/*
 * Get a "line" of sound from the soundcard.
 */
int get_sound() {
    soundDevice->read();
    
    // keep an unprocessed copy of the sound data
    memcpy(sound_data_unproc, sound_data, sound_bsize*2 );

    return 0;
}

/*****************************************************************************/


int sound_convert(unsigned char * data, int stereo, int bps) {
    unsigned int  * data_i = (unsigned int*)data;
    int i;

    if( (bps == 8) && (stereo == 2) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][1] = (int)(*data ++) - 128;
	    sound_data[i][0] = (int)(*data ++) - 128;
	}
    } else if( (bps == 8) && (stereo == 1) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][0] = (int)(*data) - 128;
	    sound_data[i][1] = (int)(*data ++) - 128;
	}	
    } else if( (bps == 16) && (stereo == 2) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][1] = (int)((*data_i ++) >> 8) - 128;
	    sound_data[i][0] = (int)((*data_i ++) >> 8) - 128;
	}
    } else if( (bps == 16) && (stereo == 1) ) {
	for(i=0; i < sound_bsize; i++) {
	    sound_data[i][1] = (int)((*data_i) >> 8) - 128;
	    sound_data[i][0] = (int)((*data_i ++) >> 8) - 128;
	}
    } 
    return 0;
}





