#include "cthugha.h"
#include "cthugha-X11.h"
#include "information.h"
#include "sound.h"
#include "action.h"
#include "translate.h"
#include "options.h"
#include "keys.h"
#include "cd_player.h"
#include "display.h"
#include "disp-sys.h"
#include "interface.h"
#include "waves.h"
#include "dga.h"

#include <unistd.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xatom.h> 
#include <signal.h>

#include "vroot.h"

#if WITH_SAVER == 1
   #include <X11/extensions/scrnsaver.h>
#endif


int xcth_saver = 0;			/* run as screen saver or not */
int xcth_panel = 1;			/* use control panel */
int xcth_dga = 0;			/* use DGA */
int xcth_sigsaver = 0;			// screen saver based on SIGCONT and SIGTSTP
xy window_pos = {0,0};
int window_do_pos = 0;
int full_screen = 0;

Display * xcth_display;
Visual * xcth_visual;
Window xcth_window;			/* window to draw into */
int xcth_screen;
int xcth_planes;
GC xcth_gc;
Colormap xcth_cmap;

int rev_byte_order;

int need_map = 0;

static Window xcth_root;		/* root window */

static Widget xcth_toplevel;
static XtAppContext xcth_app_con;

char * cthugha_mode_text() {
    return (xcth_saver || xcth_sigsaver) ? "xsav" : "xwin";
}

static void CreateWindow(char * name);


static int ignoreError(Display * /*d*/, XErrorEvent * /*e*/) {
    return 0;
}


static void resize(int new_width, int new_height) {
    if ( ! display_on_root ) {
	if( (window_size.x != new_width) || (window_size.y != new_height) ) {
	    window_size.x = new_width;
	    window_size.y = new_height;

	    XClearWindow(xcth_display, xcth_window);

	    resize_display(window_size.x, window_size.y); 
	}
    }
}

/*
 * Handlers for Buttons
 */
static void quit(int /*dummy*/ ) {
    cthugha_close = 1;

    // remove the /tmp/cthugha.pid if we are a sigsaver
    if(xcth_sigsaver)
	unlink("/tmp/cthugha.pid");

    exit(0);
}
static void key_button(Widget /*w*/, XtPointer data, XtPointer /*data2*/) {
    if(data)
	keys_x11((char*)data);
}

/*
 * create the menu buttons
 * handler for activated menu-items
 */
typedef struct {
    OptionList * opt;
    int pos;
} menu_data_t;

static void menuCB(Widget /*item*/, XtPointer data, XtPointer /*data2*/) {
    menu_data_t * d = (menu_data_t*)data;

    if(d->opt != 0)
	d->opt->changeTo(d->pos);
}

static Widget add_menu(char * name, OptionList * what, Widget parent,
		       Widget under, Widget right) {
    Arg wargs[3]; int n;
    Widget button, menu, item;
    int i;

    if(parent == NULL)
	return NULL;

    /* create menu button */
    n = 0;
    XtSetArg(wargs[n], XtNlabel, name); n++;
    if(under) {
	XtSetArg(wargs[n], XtNfromVert, under); n++;
    }
    if(right) {
	XtSetArg(wargs[n], XtNfromHoriz, right); n++;
    }
    button = XtCreateManagedWidget(name, menuButtonWidgetClass, parent, wargs, n);
    if(button == NULL)
	return NULL;

    /* create the menu popup */
    menu = XtCreatePopupShell("menu", simpleMenuWidgetClass,
			      button, NULL, 0);
    if(menu == NULL) {
	XtDestroyWidget(button);
	return NULL;
    }

    if (what->getNEntries() <= 0) {
	XtSetArg(wargs[0], XtNlabel, "none");
	XtCreateManagedWidget("none", smeBSBObjectClass, menu, wargs, 1);
    }
    
    /* create menu items */
    for(i=0; i < what->getNEntries(); i++) {
	menu_data_t * md = new menu_data_t;

	md->opt = what;
	md->pos = i;

	XtSetArg(wargs[0], XtNlabel, (*what)[i]->Name());
	item = XtCreateManagedWidget((*what)[i]->Name(), smeBSBObjectClass, menu, wargs, 1);
	XtAddCallback(item, XtNcallback, menuCB, md );
    }
    return button;
}


/*
 * Create the panel with buttons and menus 
 */
static int xcth_create_panel() {
    Widget panel;
    Widget quit_button, change_button;
    Widget menu[7];
    Arg wargs[3];

    /* create the panel formWidget */
    panel = XtCreateManagedWidget("panel", formWidgetClass, xcth_toplevel, NULL, 0);

    /* create the quit button */
    XtSetArg(wargs[0], XtNlabel, "Quit!");
    quit_button = XtCreateManagedWidget("quit", commandWidgetClass, panel, wargs, 1);
    XtAddCallback(quit_button, XtNcallback, (XtCallbackProc)quit, NULL);

    /* create the change button */
    XtSetArg(wargs[0], XtNlabel, "Change!");
    XtSetArg(wargs[1], XtNfromHoriz, quit_button);
    change_button = XtCreateManagedWidget("change", commandWidgetClass, panel, wargs, 2);
    XtAddCallback(change_button, XtNcallback, key_button, " ");
    
    /* create the menus */
    menu[0] = add_menu("Display", &screen, panel, quit_button, NULL); 
    menu[1] = add_menu("Wave", &wave,	    panel, quit_button, menu[0]);
    menu[2] = add_menu("Flame", &flame,    panel, quit_button, menu[1]);
    menu[3] = add_menu("Translation", &translate, panel, quit_button, menu[2]);
    menu[4] = add_menu("Palette", &palette,panel, quit_button, menu[3]);
    menu[5] = add_menu("PCX", &pcx,        panel, quit_button, menu[4]);
    menu[6] = add_menu("Objects", &object, panel, quit_button, menu[5]);

    /* realize window */
    XtRealizeWidget(xcth_toplevel);

    return 0;
}

/*
 * create an empty cursor
 */
static Cursor xcth_cursor() {
    Pixmap blank_pix;
    XColor dummyColor;
    GC   bit_1_gc, bit_0_gc;

    blank_pix = XCreatePixmap(xcth_display, xcth_root, 1, 1, 1);
    bit_0_gc = XCreateGC(xcth_display, blank_pix, 0, 0);
    XSetForeground(xcth_display, bit_0_gc, 0);
    
    bit_1_gc = XCreateGC(xcth_display, blank_pix, 0, 0);
    XSetForeground(xcth_display, bit_1_gc, ~0);
    
    XFillRectangle(xcth_display, blank_pix, bit_0_gc, 0, 0, 1, 1);
    
    return XCreatePixmapCursor(xcth_display, blank_pix, blank_pix, 
			       &dummyColor, &dummyColor, 0, 0);
}


/*
 * initialization for xcthugha
 */
int cth_init(int * argc, char *argv[]) {
    unsigned int byte_order_test = 0x04030201;

    seteuid(getuid());			/* give up root privileges */

#if HAVE_NCURSES == 1
    ncurses_use = text_on_term;
#else
    ncurses_use = 0;
    text_on_term = 0;
#endif

    /* intialize application */
    xcth_toplevel = XtAppInitialize(&xcth_app_con, 
				    xcth_saver ? "Cthugha-saver" : "Cthugha", 
				    NULL, 0,
				    argc, argv, NULL, NULL, 0);
    
    xcth_display = XtDisplay(xcth_toplevel);
    xcth_screen	 = DefaultScreen(xcth_display);
    xcth_visual	 = DefaultVisual(xcth_display, xcth_screen);
    xcth_planes	 = DefaultDepth(xcth_display, xcth_screen );
    bypp         = (xcth_planes+7) / 8;
    xcth_root    = DefaultRootWindow(xcth_display);

    rev_byte_order = (ImageByteOrder(xcth_display) == MSBFirst) ? 1 : 0;
    if( *(char*)&byte_order_test == 4)
	rev_byte_order = !rev_byte_order;

    return 0;
}


#if WITH_SAVER == 1

/*
 * Screen saver:
 *
 * Because of problems with XInstallColormap (only WMs should use it)
 * I use the following method to get the correct colors:
 *  - Install the screen saver, completely normal
 *  - On activation: turn off the screen saver, create a window filling the whole
 *                   screen and use that for drawing. For that window I can
 *                   specify a private colormap, so everything works find.
 */
static int  ss_event = 0;
static int  screen_saved = False;

static void StartSaver();
static void StopSaver();

/*
 * Initialize the screen saver
 */
static int init_saver() {
    XSetWindowAttributes attr;
    int			ss_error = 0;
    XID			kill_id;
    Atom		kill_type;
    XScreenSaverInfo    *info;
    unsigned long	mask;
    Pixmap		blank_pix;
    GC			black_gc;
    unsigned long	black_pixel = 0;

    exit_sound();		/* release sound again, get only when needed */

    full_screen = 1;
    window_size.x = DisplayWidth(xcth_display, xcth_screen);
    window_size.y = DisplayHeight(xcth_display, xcth_screen);

    /* check, if screen saver is possible */
    if (!XScreenSaverQueryExtension (xcth_display, &ss_event, &ss_error))
	return 1;

    XScreenSaverUnregister(xcth_display, xcth_screen);
    if(XScreenSaverGetRegistered(xcth_display,xcth_screen,&kill_id,&kill_type))
	XKillClient(xcth_display, kill_id);
    
    /* make sure, we get screen saver events */
    XScreenSaverSelectInput(xcth_display, xcth_root, ScreenSaverNotifyMask);

    blank_pix = XCreatePixmap (xcth_display, xcth_root, 1, 1, 1);
    XScreenSaverRegister(xcth_display, xcth_screen, (XID) blank_pix, XA_PIXMAP); 
    
    info = XScreenSaverAllocInfo();
    XScreenSaverQueryInfo(xcth_display, xcth_root, info);

    mask = 0;
    XScreenSaverSetAttributes(xcth_display, xcth_root, 0, 0,
			      window_size.x, window_size.y, 0,
			      CopyFromParent, CopyFromParent, 
			      CopyFromParent, mask, &attr);
			     
    XSync (xcth_display, False);

    XSetErrorHandler (ignoreError);

    black_gc = XCreateGC (xcth_display, xcth_root, 0, 0);
    XSetForeground (xcth_display, black_gc, black_pixel);

    signal(SIGINT, quit);    signal(SIGQUIT, quit);
    signal(SIGHUP, quit);    signal(SIGKILL, quit);
    signal(SIGTERM, quit);

    // create the window, it will be mapped in init_graph_mode_sys
    CreateWindow("xcthugha-saver");
	
    printfv(0, "Ready.\n");

    if (info->state == ScreenSaverOn) {
	if (info->kind != ScreenSaverExternal) {
	    XResetScreenSaver (xcth_display);
	    XActivateScreenSaver (xcth_display);
	}
	StartSaver();
    }

    return 0;
}


int ss_timeout, ss_interval, ss_prefer_blanking, ss_allow_exposures;
static void StartSaver() {
    Pixmap		blank_pix;

    if (screen_saved)			/* check if running */
	return;

    screen_saved = 2;			/* currently starting */

    /* deactivate screen saver, but remember all values */
    XGetScreenSaver(xcth_display, &ss_timeout, &ss_interval,
		    &ss_prefer_blanking, &ss_allow_exposures);
    XResetScreenSaver (xcth_display);
    XSetScreenSaver(xcth_display, 0, ss_interval,
		    ss_prefer_blanking, ss_allow_exposures);

    blank_pix = XCreatePixmap (xcth_display, xcth_root, 1, 1, 1);
    XScreenSaverRegister(xcth_display, xcth_screen, (XID) blank_pix, XA_PIXMAP); 

    XSync(xcth_display, True);

    init_graph_mode();			/* some extra initialization stuff */
    init_sound();			/* get sound only when needd */

    screen_saved = 1;
}

static void StopSaver() {
    if( screen_saved == 2)		/* currently starting */
	return;

    if (!screen_saved)			/* not saving */
	return;

    /* OK. Saver is running, now stop it */
    exit_graph_mode();
    exit_sound();

    /* reactivate saver */
    XSetScreenSaver(xcth_display, ss_timeout, ss_interval,
		    ss_prefer_blanking, ss_allow_exposures);
    screen_saved = 0;			/* no longer saving now */
}

#endif	/* WITH_SAVER */


/*
 * create a window
 */
static void CreateWindow(char * name) {
    XSetWindowAttributes attr;
    unsigned long mask;
    XSizeHints  * sh;
    unsigned long	black_pixel = 0;

    if(full_screen) {
	window_size.x = DisplayWidth(xcth_display, xcth_screen);
	window_size.y = DisplayHeight(xcth_display, xcth_screen);
    }

    mask = CWEventMask | CWCursor | CWBackPixel;
    attr.background_pixel = black_pixel;
    attr.event_mask = ButtonPressMask | PointerMotionMask | KeyReleaseMask;
    attr.cursor = xcth_cursor();

    // we can not use override_redirect, because then the window manager
    // will no install our colormap and installing the colomap ourselfs
    // with XInstallColormap is not allowed and some window manager (fvwm)
    // will set the colormap back again.
    
    /* Create the window */
    xcth_window = XCreateWindow(xcth_display, 
				DefaultRootWindow(xcth_display), 
				window_do_pos ? window_pos.x : 0, 
				window_do_pos ? window_pos.y : 0,
				window_size.x, window_size.y, 
				0, CopyFromParent, InputOutput,
				CopyFromParent, mask, &attr);
    need_map = 1;

    XStoreName(xcth_display, xcth_window, name);

    /* allocate the colormap */
    init_palettes();
    
    if(full_screen) {
	/* make sure our window get's placed automatically, and filles the whole screen */
	sh = XAllocSizeHints();
	sh->flags = USPosition | PPosition | PBaseSize;
	sh->x = sh->y = 0;
	sh->base_width = window_size.x;
	sh->base_height = window_size.y;
	XSetWMNormalHints(xcth_display, xcth_window, sh);

    } else if(window_do_pos) {
	sh = XAllocSizeHints();
	sh->flags = USPosition | PBaseSize | PPosition;
	sh->x = window_pos.x;
	sh->y = window_pos.y;
	sh->base_width = window_size.x;
	sh->base_height = window_size.y;
	XSetWMNormalHints(xcth_display, xcth_window, sh);
    } 

    XWindowAttributes wa;
    XGetWindowAttributes(xcth_display, xcth_window, &wa);
    xcth_visual	 = wa.visual;
    xcth_planes	 = wa.depth;
    bypp         = (xcth_planes+7) / 8;
    xcth_root    = wa.root;
}

/*
 *  main for xcthugha
 */
int cth_main() {
    printfv(1, "Setting X11...\n");

    /* run as screen saver */
    if(xcth_saver) {
#if WITH_SAVER == 1
	init_saver();
#else
	printfe("Screensaver was disabled at compile.\n");
	return 1;
#endif
    }

    if(xcth_sigsaver) {
	exit_sound();				/* release sound again, get only when needed */

	extern void sig_tty_cont(int);
	signal(SIGCONT, sig_tty_cont);		/* set, how to continue */

	// create the window, it will be mapped when SIGCONT arrives
	full_screen = 1;
	CreateWindow("xcthugha-saver");
	
	printfv(1, "Ready. PID: %d\n", getpid());

	char str[512];
	sprintf(str, "echo %d > /tmp/cthugha.pid", getpid());
	system(str);

	signal(SIGTSTP, SIG_DFL);
	raise(SIGTSTP);				// stop for now

	cthugha_pause = 0;

	signal(SIGINT, quit);    signal(SIGQUIT, quit);
	signal(SIGHUP, quit);    signal(SIGKILL, quit);
	signal(SIGTERM, quit);
    }

    /* with DGA support */
    if(xcth_dga) {
#if WITH_DGA == 1
	if(InitDGAEnvironment()) {
	    printfe("Unable to init DGA environment!\n");
	    return 1;
	}

	full_screen = 1;		// DGA is full screen

	if(!xcth_saver && !xcth_sigsaver) {
	    CreateWindow("xcthugha-dga");
	}
#else
	printfe("DGA was disabled at compile.\n");
	return 1;
#endif
    } 

    /* nothing special */
    if(!xcth_dga && !xcth_saver && !xcth_sigsaver) {
		     
	if(xcth_panel) {			/* create the control panel */
	    xcth_create_panel(); 
	    XSelectInput(xcth_display, XtWindow(xcth_toplevel), 
			 KeyReleaseMask | StructureNotifyMask); 
	}

	if(display_on_root) {			/* display on root window */
	    xcth_window = xcth_root;

	    XWindowAttributes wa;
	    XGetWindowAttributes(xcth_display, xcth_window, &wa);
	    xcth_visual	 = wa.visual;
	    xcth_planes	 = wa.depth;
	    bypp         = (xcth_planes+7) / 8;
	    
	    /* get size of screen */
	    window_size.x = DisplayWidth(xcth_display, xcth_screen);
	    window_size.y = DisplayHeight(xcth_display, xcth_screen);
	    
	    init_palettes();
	} else {				/* display in a window */
	    window_size = disp_size;
	    CreateWindow("Cthugha");
	}
    }

    xcth_gc = XCreateGC(xcth_display, xcth_window, 0, 0);

    if(!xcth_saver && !xcth_sigsaver)
	if(init_graph_mode())			// this also maps the window
	    return 1;

    while(cthugha_close == 0) {
	XWindowAttributes wa;
	XEvent event;
	
	while( XtAppPending(xcth_app_con) 
#if WITH_SAVER == 1
	       || (xcth_saver && !screen_saved)
#endif
	    ) {

	    XtAppNextEvent(xcth_app_con, &event);

	    if( event.type == KeyRelease) { 		/* handle keyboard input */
		char key_buff[256];
		int count;
		KeySym ks;
		XKeyEvent * kevent = (XKeyEvent*)&event;

		count = XLookupString( kevent,  key_buff, 256, &ks, NULL);
		key_buff[count] = '\0';
		if( count == 0) {
		    char * tmp = XKeysymToString(ks);
		    strncpy(key_buff, tmp ? tmp : "", 256);
		}

		/* Stop the screen saver on 'q', but don't close xcthugha */
		if( ((key_buff[0] == 'q') || (key_buff[1] == 'Q')) &&
		    (key_buff[1] == '\0') &&
		    (xcth_saver || xcth_sigsaver) ) {
#if WITH_SAVER == 1
		    if(xcth_saver)
			StopSaver();
#endif
		    if(xcth_sigsaver)
			raise(SIGTSTP);
		} else {
		    // handle key normally
		    keys_x11(key_buff, kevent->state);
		}
		
#if WITH_SAVER == 1
	    } else if (event.type == ss_event) {	/* handle screen saver events */

		XScreenSaverNotifyEvent * sevent;
		sevent = (XScreenSaverNotifyEvent *) &event;

		/* turn on screen saver */
		if (sevent->state == ScreenSaverOn) {
		    if (sevent->kind != ScreenSaverExternal) {
			XResetScreenSaver (xcth_display);
			XActivateScreenSaver (xcth_display);
		    } else {
			StartSaver ();
		    }
		    /* turn off screen saver */
		} else if (sevent->state == ScreenSaverOff) {
		    StopSaver ();
		}
	    } else if( (event.type == MotionNotify) && xcth_saver ) {
		StopSaver();
#endif
	    } else if( (event.type == MotionNotify) && xcth_sigsaver ) {
		raise(SIGTSTP);
		
	    } else {
		XtDispatchEvent(&event);
	    }

	    currentInterface->run();
	}
    	XGetWindowAttributes(xcth_display, xcth_window, &wa);
	resize(wa.width, wa.height); 

	display_sound(); 
	currentInterface->run();
    }

    return 0;
}
    












